from .Response import Response
from .Value import Value

from collections import OrderedDict
import logging
from threading import Lock


class Node:
    """
    Represents a Node on the Node structure.
    """

    def __init__(self, name, parent, standalone=False):
        """
        Node Constructor.
        :param name: Node name.
        :param parent: Node parent.
        :param standalone: Standalone Node structure.
        """
        self.logger = logging.getLogger("DSLink")
        if parent is not None:
            self.link = parent.link
        self.parent = parent
        self.standalone = standalone
        self.transient = False
        self.value = Value()
        self.children = {}
        self.children_lock = Lock()
        self.config = OrderedDict([("$is", "node")])
        self.attributes = OrderedDict()
        self.streams = []
        self.removed_children = []
        self.removed_children_lock = Lock()
        # TODO(logangorence): Deprecate for v0.6
        self.set_value_callback = None
        # TODO(logangorence): Normalize path?
        if parent is not None:
            self.name = name
            if parent.path.endswith("/"):
                self.path = parent.path + name
            else:
                self.path = parent.path + "/" + name
        else:
            if name is not None:
                self.name = name
                self.path = "/" + name
            else:
                self.name = ""
                self.path = ""

    def get_type(self):
        """
        Get the Node's value type.
        :return: Value type.
        """
        return self.get_config("$type")

    def set_type(self, t):
        """
        Set the Node's value type.
        :param t: Type to set.
        """
        self.value.set_type(t)
        self.set_config("$type", t)

    def get_value(self):
        """
        Get the Node's value.
        :return: Node value.
        """
        if self.value.type is "enum":
            return Value.build_enum(self.value.value)
        return self.value.value

    def set_value(self, value, trigger_callback=False, check=True):
        """
        Set the Node's value.
        :param value: Value to set.
        :param trigger_callback: Set to true if you want to trigger the value set callback.
        :param check: Turn type checking off if false.
        :return: True if the value was set.
        """
        # Set value and updated timestamp
        i = self.value.set_value(value, check)
        if i and (not self.standalone or self.link_is_active()):
            self.nodes_changed()
            self.update_subscribers_values()
            if trigger_callback:
                if hasattr(self.set_value_callback, "__call__"):
                    self.set_value_callback(node=self, value=value)
                try:
                    self.link.responder.profile_manager.get_profile(self.get_config("$is")).run_set_callback((self, value))
                except ValueError:
                    pass
        return i

    def get_config(self, key):
        """
        Get a config value.
        :param key: Key of config.
        :return: Value of config.
        """
        return self.config[key]

    def set_config(self, key, value):
        """
        Set a config value.
        :param key: Key of config.
        :param value: Value of config.
        """
        self.nodes_changed()
        self.config[key] = value
        self.update_subscribers()

    def get_attribute(self, key):
        """
        Get an attribute value.
        :param key: Key of attribute.
        :return: Value of attribute.
        """
        return self.attributes[key]

    def set_attribute(self, key, value):
        """
        Set an attribute value.
        :param key: Key of attribute.
        :param value: Value of attribute.
        """
        self.nodes_changed()
        self.attributes[key] = value
        self.update_subscribers()

    def set_transient(self, transient):
        """
        Set the node to be transient, which won't serialize it.
        :param transient: True if transient.
        """
        if type(transient) is not bool:
            raise TypeError("Transient must be bool")
        self.transient = transient

    def nodes_changed(self):
        """
        Set the Node structure as changed.
        """
        self.link.responder.nodes_changed = True

    def link_is_active(self):
        """
        Check if the link is active.
        :return: True if link is active.
        """
        return self.link.active

    def set_display_name(self, name):
        """
        Set the Node name.
        :param name: Node name.
        """
        if not isinstance(name, basestring):
            raise ValueError("Passed profile is not a string")
        self.set_config("$name", name)
        self.update_subscribers()

    def set_invokable(self, invokable):
        """
        Set invokable state.
        :param invokable: Invokable permit string or true for everyone can access.
        """
        if isinstance(invokable, basestring):
            self.set_config("$invokable", invokable)
        elif type(invokable) is bool and invokable:
            self.set_config("$invokable", "read")
        else:
            raise ValueError("Invokable is not a string or boolean")

    def set_parameters(self, parameters):
        """
        Set parameters for action.
        :param parameters: Parameters for action.
        """
        if type(parameters) is not list:
            raise ValueError("Parameters is not a list")
        self.set_config("$params", parameters)

    def set_columns(self, columns):
        """
        Set return columns for action.
        :param columns: Columns for action.
        """
        if type(columns) is not list:
            raise ValueError("Columns is not a list")
        self.set_config("$columns", columns)

    def set_profile(self, profile):
        """
        Set the Node's profile.
        :param profile: Node Profile.
        """
        if not isinstance(profile, basestring):
            raise ValueError("Passed profile is not a string")
        self.set_config("$is", profile)

    def set_writable(self, permission):
        """
        Set the writable permission.
        :param permission: Permission to set.
        """
        if isinstance(permission, basestring):
            self.set_config("$writable", permission)
        else:
            raise ValueError("Passed permission is not string")

    def stream(self):
        """
        Stream the Node.
        :return: Node stream.
        """
        out = []
        for key in self.config:
            value = self.config[key]
            if key == "$$password":
                value = None
            out.append([key, value])
        for key in self.attributes:
            value = self.attributes[key]
            out.append([key, value])
        with self.children_lock:
            for child in self.children:
                child = self.children[child]
                if child.value.has_value():
                    val = {
                        "value": child.value.value,
                        "ts": child.value.updated_at.isoformat()
                    }
                else:
                    val = {}
                child_data = dict(child.config)
                child_data.update(child.attributes)
                child_data.update(val)
                if "$$password" in child_data:
                    child_data["$$password"] = None
                out.append([
                    child.name,
                    child_data
                ])
        with self.removed_children_lock:
            for child in self.removed_children:
                out.append({
                    "name": child.name,
                    "change": "remove"
                })
                self.nodes_changed()
            del self.removed_children[:]
        return out

    def add_child(self, child):
        """
        Add a child to this Node.
        :param child: Child to add.
        """
        with self.children_lock:
            if child.name in self.children:
                raise ValueError("Child %s already exists in %s" % (child.name, self.path))
            self.children[child.name] = child
        self.nodes_changed()

        if self.standalone or self.link_is_active():
            self.update_subscribers()

    def remove_child(self, name):
        """
        Remove a child from this Node.
        :param name: Child Node name.
        """
        if name not in self.children:
            return
        with self.children_lock:
            self.removed_children.append(self.children.pop(name))
        self.update_subscribers()

    def has_child(self, name):
        """
        Check if this Node has child of name.
        :param name: Name of child.
        :return: True if the child of name exists.
        """
        return name in self.children

    def create_child(self, name):
        """
        Create a child node that inherits the root
        link and parent is this Node.
        :param name: Name of child node to create.
        :return: Child node.
        """
        child = Node(name, self)
        self.add_child(child)
        return child

    def get(self, path):
        """
        Get a Node from this position on the Node structure.
        :param path: Path of Node wanted.
        :return: Node of path.
        """
        if path == "/":
            return self
        elif path.startswith("/$"):
            return self
        elif path.startswith("/@"):
            return self
        else:
            try:
                try:
                    i = path.index("/", 2)
                    child = path[1:i]
                    return self.children[child].get(path[i:])
                except ValueError:
                    child = path[1:]
                    try:
                        return self.children[child]
                    except KeyError:
                        return None
            except KeyError:
                import traceback
                self.logger.warn("Non-existent Node requested %s on %s" % (path, self.path))

    def set_config_attr(self, path, value):
        """
        Set value/config/attribute on Node.
        :param path: Path of value to set.
        :param value: Value to set.
        """
        if path == "/" or path == self.path:
            self.set_value(value, trigger_callback=True)
        elif path.startswith("/$") or path.startswith(self.path + "/$"):
            self.set_config(path[2:], value)
        elif path.startswith("/@") or path.startswith(self.path + "/@"):
            self.set_attribute(path[2:], value)
        else:
            node = self.get(path)
            if node is not None:
                node.set_config_attr(path, value)

    def remove_config_attr(self, path):
        """
        Remove config/attribute on Node.
        :param path: Path of value to remove.
        """
        if path.startswith("/$") or path.startswith(self.path + "/$"):
            del self.config[path[2:]]
        elif path.startswith("/@") or path.startswith(self.path + "/@"):
            del self.config[path[2:]]
        else:
            self.get(path).remove_config_attr(path)

    def is_subscribed(self):
        """
        Check whether the Node is subscribed to.
        :return: True if the Node is subscribed to.
        """
        sub = self.link.responder.subscription_manager.get_sub(self.path)
        if sub is None:
            return False
        return len(sub.sid_qos) is not 0

    def invoke(self, params):
        """
        Invoke the Node.
        :param params: Parameters of invoke.
        :return: Columns and values
        """
        self.logger.debug("%s invoked, with parameters: %s" % (self.path, params))
        try:
            # noinspection PyCallingNonCallable
            return (self.config["$columns"] if "$columns" in self.config else []), self.link.responder.profile_manager.get_profile(self.get_config("$is")).run_callback((self, params))
        except ValueError:
            return [], []

    def update_subscribers(self):
        """
        Send subscription updates.
        """
        responses = []
        for stream in self.streams:
            responses.append(Response({
                "rid": stream,
                "stream": "open",
                "updates": self.stream()
            }).get_stream())
        if responses:
            self.link.wsp.sendMessage({
                "responses": responses
            })

    def update_subscribers_values(self):
        """
        Update all Subscribers of a Value change.
        """
        if self.value.has_value():
            self.link.responder.subscription_manager.send_value_update(self)

    def to_json(self):
        """
        Convert to an object that is saved to JSON.
        :return: JSON object.
        """
        out = {}

        for key in self.config:
            out[key] = self.config[key]
        for key in self.attributes:
            out[key] = self.attributes[key]
        if self.value.has_value():
            out["?value"] = self.value.value
        for child in self.children:
            if not self.children[child].transient:
                out[child] = self.children[child].to_json()

        return out

    @staticmethod
    def from_json(obj, root, name, link=None):
        """
        Convert a JSON object to a String
        :param obj: Node Object.
        :param root: Root Node.
        :param name: Node Name.
        :param link: Created Node's link.
        :return: Node that was created.
        """
        node = Node(name, root)
        if link is not None:
            node.link = link

        if type(obj) is dict:
            for prop in obj:
                if prop.startswith("$"):
                    if prop == "$type":
                        node.set_type(obj[prop])
                    else:
                        node.set_config(prop, obj[prop])
                elif prop.startswith("@"):
                    node.set_attribute(prop, obj[prop])
                elif prop == "?value":
                    node.set_value(obj[prop], check=False)
                else:
                    node.add_child(Node.from_json(obj[prop], node, prop))

        return node

    @staticmethod
    def normalize_path(path, leading):
        """
        :param path: Path to normalize.
        :param leading: True if leading forward slash is kept/added, false if not.
        :return: Normalized path.
        """
        if not leading and path.startswith("/"):
            path = path[1:]
        elif leading and not path.startswith("/"):
            path = "/" + path
        if path.endswith("/"):
            path = path[:-1]

        return path


class RemoteNode(Node):
    def __init__(self, name, parent, parent_path=None):
        Node.__init__(self, name, parent)
        if parent_path is not None:
            self.path = parent_path + name

    def set_value(self, value, trigger_callback=False, check=True):
        # TODO(logangorence): Set value.
        pass

    def update_subscribers(self):
        pass

    def update_subscribers_values(self):
        pass

    def add_child(self, child):
        pass

    def nodes_changed(self):
        pass

    def link_is_active(self):
        return False

    def from_serialized(self, serialized):
        for a in serialized:
            if type(a) is not list:
                continue
            k = a[0]
            v = a[1]
            if k.startswith("$"):
                self.set_config(k, v)
            elif k.startswith("@"):
                self.set_attribute(k, v)
            else:
                child = RemoteNode(k, self)
                for i in v:
                    if i.startswith("$"):
                        child.set_config(i, v[i])
                    elif i.startswith("@"):
                        child.set_attribute(i, v[i])
                Node.add_child(self, child)



import argparse
import base64
import hashlib
import logging
import signal
from twisted.internet import reactor
try:
    from urlparse import urlparse
except ImportError:
    from urllib import parse as urlparse

from .FileStorage import FileStorage
from .Crypto import Keypair
from .Handshake import Handshake
from .Requester import Requester
from .Responder import Responder
from .WebSocket import WebSocket


class DSLink:
    """
    Base DSLink class which creates the node structure,
    subscription/stream manager, and connects to the broker.
    """

    def __init__(self, config):
        """
        Construct for DSLink.
        :param config: Configuration object.
        """
        self.active = False
        self.needs_auth = False

        # DSLink Configuration
        self.config = config
        self.server_config = None
        self.storage = FileStorage(self)

        # Logger setup
        self.logger = self.create_logger("DSLink", self.config.log_level)
        self.logger.info("Starting DSLink")

        # Signal setup
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)

        # Requester and Responder setup
        if self.config.requester:
            self.requester = Requester(self)
        if self.config.responder:
            self.responder = Responder(self)
            self.responder.start()

        # DSLink setup
        self.keypair = Keypair(self.config.keypair_path)
        self.handshake = Handshake(self, self.keypair)
        self.handshake.run_handshake()
        self.dsid = self.handshake.get_dsid()

        # Connection setup
        self.wsp = None
        self.websocket = WebSocket(self)

        self.call_later(1, self.start)

        self.logger.info("Started DSLink")
        self.logger.debug("Starting reactor")
        reactor.run(installSignalHandlers=False)

    def start(self):
        """
        Called once the DSLink is initialized and connected.
        Override this rather than the constructor.
        """
        pass

    # noinspection PyUnresolvedReferences
    def stop(self, *args):
        """
        Called when the DSLink is going to stop.
        Override this if you start any threads you need to stop.
        Be sure to call the super function.
        :param args: Signal arguments.
        """
        if self.wsp is not None:
            reactor.callFromThread(self.wsp.sendClose)
        reactor.removeAll()
        reactor.iterate()
        reactor.stop()

    # noinspection PyMethodMayBeStatic
    def get_default_nodes(self, super_root):
        """
        Create the default Node structure in this, override it.
        :param super_root: Super Root.
        :return: Super Root with default Node structure.
        """
        return super_root

    def get_auth(self):
        """
        Get auth parameter for connection.
        :return: Auth parameter.
        """
        auth = self.server_config["salt"].encode("ascii", "ignore")
        auth += self.shared_secret
        auth = base64.urlsafe_b64encode(hashlib.sha256(auth).digest()).decode("utf-8").replace("=", "")
        return auth

    def get_url(self):
        """
        Get full WebSocket URL.
        :return: WebSocket URL.
        """
        websocket_uri = self.config.broker[:-5].replace("http", "ws") + "/ws?dsId=%s" % self.dsid
        if self.needs_auth:
            websocket_uri += "&auth=%s" % self.get_auth()
        token = self.config.token_hash(self.dsid, self.config.token)
        if token is not None:
            websocket_uri += token
        url = urlparse(websocket_uri)
        if url.port is None:
            port = 80
        else:
            port = url.port
        return websocket_uri, url, port

    @staticmethod
    def create_logger(name, log_level=logging.INFO):
        """
        Create a logger with the specified name.
        :param name: Logger name.
        :param log_level: Output Logger level.
        :return: Logger instance.
        """
        # Logger setup
        formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        ch = logging.StreamHandler()
        ch.setFormatter(formatter)
        ch.setLevel(log_level)
        logger = logging.getLogger(name)
        logger.setLevel(log_level)
        logger.addHandler(ch)
        return logger

    @staticmethod
    def call_later(delay, call, *args, **kw):
        """
        Call function later.
        :param delay: Seconds to delay.
        :param call: Method to call.
        :param args: Arguments.
        :return: DelayedCall instance.
        """
        return reactor.callLater(delay, call, *args, **kw)


class Configuration:
    """
    Provides configuration to the DSLink.
    """

    def __init__(self, name, responder=False, requester=False, ping_time=30, keypair_path=".keys",
                 nodes_path="nodes.json", no_save_nodes=False):
        """
        Object that contains configuration for the DSLink.
        :param name: DSLink name.
        :param responder: True if responder, default is False.
        :param requester: True if requester, default is False.
        :param ping_time: Time between pings, default is 30.
        :param keypair_path: Path to save keypair, default is ".keys".
        :param nodes_path: Path to save nodes.json, default is "nodes.json".
        :param no_save_nodes: Don't use nodes.json, default is False.
        """
        if not responder and not requester:
            raise ValueError("DSLink is neither responder nor requester.")
        parser = argparse.ArgumentParser()
        parser.add_argument("--broker", default="http://localhost:8080/conn")
        parser.add_argument("--log", default="info")
        parser.add_argument("--token")
        args = parser.parse_args()
        self.name = name
        self.dsid = None
        self.broker = args.broker
        self.log_level = args.log.lower()
        self.token = args.token
        self.responder = responder
        self.requester = requester
        self.ping_time = ping_time
        self.keypair_path = keypair_path
        self.nodes_path = nodes_path
        self.no_save_nodes = no_save_nodes

        if self.log_level == "critical":
            self.log_level = logging.CRITICAL
        elif self.log_level == "error":
            self.log_level = logging.ERROR
        elif self.log_level == "warning":
            self.log_level = logging.WARNING
        elif self.log_level == "info":
            self.log_level = logging.INFO
        elif self.log_level == "debug":
            self.log_level = logging.DEBUG
        elif self.log_level == "none":
            self.log_level = logging.NOTSET

    @staticmethod
    def token_hash(dsid, token):
        if token is not None and len(token) > 16:
            token_id = token[0:16]
            hash_str = base64.urlsafe_b64encode(hashlib.sha256((dsid + token)).digest()).decode("utf-8").replace("=", "")
            return "&token=" + token_id + hash_str
        else:
            return None



class StorageDriver:
    def __init__(self):
        pass

    def read(self):
        """
        Read from storage.
        """
        pass

    def store(self, subscription, value):
        """
        Store value in storage.
        """
        pass

    def get_updates(self, path, sid):
        pass



import copy
import os
import pickle

from .Storage import StorageDriver
from .Util import *
from .Value import Value


class FileStorage(StorageDriver):
    def __init__(self, link, path="storage/"):
        StorageDriver.__init__(self)
        self.link = link
        self.storage = path
        self.updates_cache = {}
        self.update_cache = {}

    def read(self):
        ret = {}
        if not os.path.isdir(self.storage):
            return ret
        files = os.listdir(self.storage)
        if len(files) is 0:
            return ret
        for f in files:
            file = open(self.storage + "/" + f, "rb")
            json_obj = pickle.load(file)
            file.close()
            qos = json_obj["qos"]
            path = base64_decode(f)
            sub = {
                "path": path,
                "qos": qos
            }
            ret[path] = sub
            if qos is 2:
                value = Value()
                value.type = json_obj["type"]
                value.updated_at = json_obj["ts"]
                value.value = json_obj["value"]
                self.store(sub, value)
                self.update_cache[path] = value
            elif qos is 3:
                try:
                    import_queue = json_obj["queue"]
                except KeyError:
                    continue

                if path not in self.updates_cache or sub["path"] not in self.updates_cache:
                    queue = []
                elif path in self.updates_cache:
                    queue = self.updates_cache[path]
                elif sub["path"] in self.updates_cache:
                    queue = self.updates_cache[sub["path"]]
                else:
                    continue

                for array in import_queue:
                    value = Value()
                    value.type = array[0]
                    value.updated_at = array[1]
                    value.value = array[2]
                    queue.append(value)
        return ret

    def store_qos2(self, subscription, value, json_obj):
        self.update_cache[subscription.path] = value
        json_obj = {
            "qos": 2
        }
        if value is not None:
            json_obj["type"] = value.type
            json_obj["ts"] = value.updated_at
            json_obj["value"] = value.value
        return json_obj

    def store_qos3(self, subscription, value, json_obj):
        if value is None:
            return json_obj
        if subscription.path in self.updates_cache:
            cache = self.updates_cache[subscription.path]
        else:
            cache = []
            self.updates_cache[subscription.path] = cache
        cache.append(copy.copy(value))
        if len(cache) > 1000:
            del cache[0]
        queue = []
        json_obj = {
            "queue": queue,
            "qos": 3
        }
        for v in cache:
            if value is None:
                queue.append(None)
            else:
                array = [
                    v.type,
                    v.updated_at,
                    v.value
                ]
                queue.append(array)
        return json_obj

    def store(self, subscription, value):
        qos2 = False
        qos3 = False
        for sid in subscription.sid_qos:
            qos = subscription.sid_qos[sid]
            if qos is 2:
                qos2 = True
            if qos is 3:
                qos3 = True
        json_obj = None
        if qos2:
            json_obj = self.store_qos2(subscription, value, json_obj)
        if qos3:
            json_obj = self.store_qos3(subscription, value, json_obj)
        if json_obj is not None:
            if not(os.path.exists(self.storage) or os.mkdir(self.storage)):
                pass
            file = open(self.storage + base64_encode(subscription.path), "wb")
            pickle.dump(json_obj, file)
            file.close()

    def get_updates(self, path, sid):
        cache = self.updates_cache.pop(path, None)
        tmp = self.updates_cache.pop(path, None)
        if tmp is not None:
            return [
                sid,
                tmp.value.value,
                tmp.value.updated_at.isoformat()
            ]
        if cache is None or len(cache) is 0:
            return None
        updates = []
        for val in cache:
            print("replaying " + str(val.value))
            update = [
                sid,
                val.value,
                val.updated_at.isoformat()
            ]
            updates.append(update)
            file_path = self.storage + base64_encode(path)
            if os.path.exists(file_path):
                os.remove(file_path)
        return updates



from __future__ import print_function
from .Profile import ProfileManager
from .Node import Node

import json
import os.path
from twisted.internet import task


class Responder:
    def __init__(self, link):
        """
        Create Responder class.
        :param link: DSLink instance.
        """
        self.link = link
        self.nodes_changed = False
        self.subscription_manager = LocalSubscriptionManager(link)
        self.stream_manager = StreamManager(link)
        self.profile_manager = ProfileManager(link)

    def start(self):
        """
        Start Responder.
        :return:
        """
        # Load or create an empty Node structure
        self.load_nodes()
        self.create_defs()

        # Start saving timer
        if not self.link.config.no_save_nodes:
            task.LoopingCall(self.save_nodes).start(5)

    def get_super_root(self):
        """
        Get Super Root.
        :return: Super Root.
        """
        return self.super_root

    def create_defs(self):
        """
        Create /defs/ Node.
        """
        defs = Node("defs", self.get_super_root())
        defs.set_transient(True)
        defs.set_config("$hidden", True)
        defs.add_child(Node("profile", defs))
        self.get_super_root().add_child(defs)

    def create_empty_super_root(self):
        """
        Create empty super root.
        :return: Empty Super Root.
        """
        super_root = Node("", None)
        super_root.link = self.link
        return super_root

    # noinspection PyBroadException
    def load_nodes(self):
        """
        Load nodes.json file from disk, use backup if necessary. If that fails, then reset to defaults.
        """
        nodes_path = self.link.config.nodes_path
        if os.path.exists(nodes_path):
            try:
                nodes_file = open(nodes_path, "r")
                obj = json.load(nodes_file)
                nodes_file.close()
                self.super_root = Node.from_json(obj, None, "", link=self.link)
            except Exception as e:
                print(e)
                self.link.logger.error("Unable to load nodes data")
                if os.path.exists(nodes_path + ".bak"):
                    try:
                        self.link.logger.warn("Restoring backup nodes")
                        os.remove(nodes_path)
                        os.rename(nodes_path + ".bak", nodes_path)
                        nodes_file = open(nodes_path, "r")
                        obj = json.load(nodes_file)
                        nodes_file.close()
                        self.super_root = Node.from_json(obj, None, "", link=self.link)
                    except:
                        self.link.logger.error("Unable to restore nodes, using default")
                        fresh = True
                else:
                    self.link.logger.warn("Backup nodes data doesn't exist, using default")
                    fresh = True
        else:
            fresh = True
        if fresh:
            self.super_root = self.create_empty_super_root()
            self.link.get_default_nodes(self.super_root)

    def save_nodes(self):
        """
        Save the nodes.json out to disk if changed, and create the bak file.
        """
        if self.nodes_changed:
            if os.path.exists(self.link.config.nodes_path + ".bak"):
                os.remove(self.link.config.nodes_path + ".bak")
            if os.path.exists(self.link.config.nodes_path):
                os.rename(self.link.config.nodes_path, self.link.config.nodes_path + ".bak")
            nodes_file = open(self.link.config.nodes_path, "w")
            nodes_file.write(json.dumps(self.get_super_root().to_json(), sort_keys=True, indent=2))
            nodes_file.flush()
            os.fsync(nodes_file.fileno())
            nodes_file.close()
            self.nodes_changed = False


# TODO: maybe rename to something more fitting?
class Subscription:
    def __init__(self, path, sid_qos):
        self.path = path
        self.sid_qos = sid_qos


class LocalSubscriptionManager:
    """
    Manages subscriptions to local Nodes.
    """

    def __init__(self, link):
        self.link = link
        self.path_subs = {}
        self.sids_path = {}
        subs = self.link.storage.read()
        for path in subs:
            sub = subs[path]
            self.path_subs[path] = Subscription(path, {-1: sub["qos"]})

    def get_sub(self, path):
        path = Node.normalize_path(path, True)
        if path not in self.path_subs:
            return None
        return self.path_subs[path]

    def add_value_sub(self, node, sid, qos=0):
        """
        Store a Subscription to a Node.
        :param node: Node to subscribe to.
        :param sid: SID of Subscription.
        :param qos: Quality of Service.
        """
        path = Node.normalize_path(node.path, True)

        if path not in self.path_subs:
            sub = Subscription(path, {sid: qos})
            self.path_subs[path] = sub
        else:
            self.path_subs[path].sid_qos[sid] = qos
            self.path_subs[path].qos = qos
        self.sids_path[sid] = path
        updates = self.link.storage.get_updates(path, sid)
        if updates is not None:
            self.link.wsp.sendMessage({
                "responses": [
                    {
                        "rid": 0,
                        "updates": updates
                    }
                ]
            })
        else:
            node.update_subscribers_values()

    def remove_value_sub(self, sid):
        """
        Remove a Subscription to a Node.
        :param sid: SID of Subscription.
        """
        if sid in self.sids_path:
            path = self.sids_path[sid]
            del self.sids_path[sid]
            if path in self.path_subs:
                del self.path_subs[path].sid_qos[sid]

    def send_value_update(self, node):
        if node.path not in self.path_subs:
            return
        msg = {
            "responses": []
        }
        for sid in self.path_subs[node.path].sid_qos:
            qos = self.path_subs[node.path].sid_qos[sid]
            if not self.link.active and qos > 0:
                self.link.storage.store(self.path_subs[node.path], node.value)
            msg["responses"].append({
                "rid": 0,
                "updates": [
                    [
                        sid,
                        node.value.value,
                        node.value.updated_at.isoformat()
                    ]
                ]
            })
        if len(msg["responses"]) is not 0:
            self.link.wsp.sendMessage(msg)


class StreamManager:
    """
    Manages streams for Nodes.
    """

    def __init__(self, link):
        """
        Constructor of StreamManager.
        """
        self.link = link
        self.streams = {}

    def open_stream(self, node, rid):
        """
        Open a Stream.
        :param node: Node to handle streaming.
        :param rid: RID of Stream.
        """
        self.streams[rid] = node
        self.streams[rid].streams.append(rid)

    def close_stream(self, rid):
        """
        Close a Stream.
        :param rid: RID of Stream.
        """
        try:
            self.streams[rid].streams.remove(rid)
            del self.streams[rid]
        except KeyError:
            self.link.logger.debug("Unknown rid %s" % rid)



from .Node import Node


class Profile:
    def __init__(self, name):
        self.name = name
        self.invoke_callback = None
        self.set_callback = None

    def run_callback(self, parameters):
        """
        Invoke the invoke callback.
        :param parameters: Parameters for callback(Node, Parameters).
        :return: Results.
        """
        if hasattr(self.invoke_callback, "__call__"):
            return self.invoke_callback(parameters)
        else:
            raise TypeError("Profile %s does not define an invoke callback" % self.name)

    def run_set_callback(self, parameters):
        """
        Invoke the set callback.
        :param parameters: Parameters for callback(Node, Value).
        :return: Results.
        """
        if hasattr(self.set_callback, "__call__"):
            return self.set_callback(parameters)
        else:
            raise TypeError("Profile %s does not define a set callback" % self.name)


class ProfileManager:
    """
    Class that holds Profiles.
    """

    def __init__(self, link):
        """
        ProfileManager Constructor.
        :param link: DSLink instance.
        """
        self.profiles = {}
        self.link = link

    def create_profile(self, profile):
        """
        Create a Profile.
        :param profile: Profile name.
        :return: Profile instance.
        """
        profile_node = self.link.responder.get_super_root().get("/defs/profile/")
        if profile in self.profiles and profile_node.has_child(profile):
            raise ValueError("Profile %s already exists" % profile)
        profile_inst = Profile(profile)
        self.profiles[profile] = profile_inst
        profile_node.add_child(Node(profile, profile_node))
        return profile_inst

    def get_profile(self, profile):
        """
        Get Profile with specified name.
        :param profile: Profile name.
        :return: Profile.
        """
        if profile not in self.profiles:
            raise ValueError("Profile %s doesn't exist" % profile)
        return self.profiles[profile]

    def register_callback(self, profile, callback):
        """
        Register a callback when the Node is invoked.
        :param profile: Profile name.
        :param callback: Callback.
        """
        if profile not in self.profiles:
            raise ValueError("Profile %s doesn't exist" % profile)
        self.profiles[profile].invoke_callback = callback

    def deregister_callback(self, profile):
        """
        Deregister an invoke callback.
        :param profile: Profile name.
        """
        if profile not in self.profiles:
            raise ValueError("Profile %s doesn't exist" % profile)
        del self.profiles[profile].invoke_callback

    def register_set_callback(self, profile, callback):
        """
        Register a callback when the Node's value is set.
        :param profile: Profile name.
        :param callback: Callback.
        """
        if profile not in self.profiles:
            raise ValueError("Profile %s doesn't exist" % profile)
        self.profiles[profile].set_callback = callback

    def deregister_set_callback(self, profile):
        """
        Deregister a set callback.
        :param profile: Profile name.
        """
        if profile not in self.profiles:
            raise ValueError("Profile %s doesn't exist" % profile)
        del self.profiles[profile].set_callback

    def get_profile_node(self, profile):
        """
        Get a profile's Node.
        :param profile: Profile name.
        :return: Profile's Node.
        """
        if profile not in self.profiles:
            raise ValueError("Profile %s doesn't exist" % profile)
        return self.link.responder.get_super_root().get("/defs/profile/%s")



import datetime

from .Response import Response

import logging


class Request:
    """
    Class that handles incoming requests.
    """

    def __init__(self, request, link):
        self.logger = logging.getLogger("DSLink")
        self.request = request
        self.link = link
        self.rid = request["rid"]
        self.method = request["method"]

    def process(self):
        """
        Process the request.
        :return: Response adequate the the request.
        """
        if self.method == "list":
            self.logger.debug("List method")
            node = self.link.responder.get_super_root().get(self.request["path"])
            resp = Response({
                "rid": self.rid,
                "stream": "open"
            })
            if node is not None:
                self.link.responder.stream_manager.open_stream(node, self.rid)
                resp.json["updates"] = node.stream()
            else:
                resp.json["updates"] = [
                    [
                        "$disconnectedTs",
                        datetime.datetime.now().isoformat()
                    ]
                ]
            return resp
        elif self.method == "subscribe":
            self.logger.debug("Subscribe method")
            for sub in self.request["paths"]:
                node = self.link.responder.get_super_root().get(sub["path"])
                if node is not None:
                    qos = 0
                    if "qos" in sub:
                        qos = sub["qos"]
                    self.link.responder.subscription_manager.add_value_sub(node, sub["sid"], qos)
                    self.logger.debug("Subscription added")
            return Response({
                "rid": self.rid,
                "stream": "closed"
            })
        elif self.method == "unsubscribe":
            for sid in self.request["sids"]:
                self.logger.debug("Unsubscribe from sid %s" % sid)
                self.link.responder.subscription_manager.remove_value_sub(sid)
            return Response({
                "rid": self.rid,
                "stream": "closed"
            })
        elif self.method == "invoke":
            self.logger.debug("Invoke method")
            # TODO(logangorence) Handle not implemented profiles
            node_invoke = self.link.responder.get_super_root().get(self.request["path"])
            if node_invoke is not None:
                columns, values = node_invoke.invoke(self.request["params"])
            else:
                self.logger.debug("Invoke on non-existant node %s." % self.request["path"])
                columns = []
                values = []
            # TODO(logangorence) Implement streaming invokes
            return Response({
                "rid": self.rid,
                "columns": columns,
                "updates": values,
                "stream": "closed"
            })
        elif self.method == "set":
            self.logger.debug("Set method")
            # TODO(logangorence) Handle permit
            # TODO(logangorence) Handle improper value type
            self.link.responder.get_super_root().set_config_attr(self.request["path"], self.request["value"])
            return Response({
                "rid": self.rid,
                "stream": "closed"
            })
        elif self.method == "remove":
            self.logger.debug("Remove method")
            self.link.responder.get_super_root().remove_config_attr(self.request["path"])
            return Response({
                "rid": self.rid,
                "stream": "closed"
            })
        elif self.method == "close":
            self.logger.debug("Close method")
            self.link.responder.stream_manager.close_stream(self.rid)
            return Response({
                "rid": self.rid,
                "stream": "closed"
            })
        else:
            raise NotImplementedError("Method %s is not implemented" % self.method)



class Response:
    def __init__(self, json):
        self.json = json
        self.rid = json["rid"]

    def get_stream(self):
        return self.json


class ListResponse(Response):
    def __init__(self, json, path):
        Response.__init__(self, json)
        self.path = path
        self.node = self.process(json["updates"])

    def process(self, update):
        from .Node import RemoteNode
        node = RemoteNode(self.path.split("/")[-1], None)
        node.path = self.path
        node.link = None
        node.from_serialized(update)
        return node


class SubscriptionResponse(Response):
    def __init__(self, json, path):
        # TODO(logangorence): Close response.
        Response.__init__(self, json)



import base64


def base64_encode(data):
    return base64.urlsafe_b64encode(data).rstrip("=")


def base64_decode(data):
    return base64.urlsafe_b64decode(base64_add_padding(data))


def base64_add_padding(string):
    """
    Add padding to a URL safe base64 string.
    :param string: Non-padded Url-safe Base64 string.
    :return: Padded Url-safe Base64 string.
    """
    while len(string) % 4 != 0:
        string += "="
    return string



from .Util import *

import json


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytearray):
            return "\x1Bbytes:" + base64_encode(obj)
        else:
            return json.JSONEncoder.default(self, obj)



from .DSLink import DSLink, Configuration
from .FileStorage import FileStorage
from .Node import Node
from .Permission import Permission
from .Profile import Profile
from .Requester import Requester
from .Response import *
from .Storage import StorageDriver
from .Value import Value

__title__ = "dslink"
__version__ = "0.6.21"
__author__ = "Logan Gorence"

__license__ = "Apache 2.0"
__copyright__ = "Copyright 2015 DGLogik Inc."



from .JsonSerializer import JsonEncoder
from .Request import Request
from .Response import Response

import json
import logging

from autobahn.twisted.websocket import WebSocketClientProtocol, WebSocketClientFactory
from autobahn.websocket.util import parse_url
from twisted.internet import reactor, task
from twisted.internet.protocol import ReconnectingClientFactory


class WebSocket:
    """
    Class to handle WebSocket.
    """

    def __init__(self, link):
        """
        WebSocket Constructor.
        :param link: DSLink instance.
        """
        websocket_uri, url, port = link.get_url()

        DSAWebSocket.link = link
        factory = DSAWebSocketFactory(websocket_uri, link)
        factory.protocol = DSAWebSocket

        link.logger.debug("Connecting WebSocket to %s" % websocket_uri)
        self.connector = reactor.connectTCP(url.hostname, port, factory)


class DSAWebSocketFactory(WebSocketClientFactory, ReconnectingClientFactory):
    def __init__(self, websocket_uri, link):
        super(DSAWebSocketFactory, self).__init__(websocket_uri)
        self.link = link
        self.cooldown = 1
    
    def clientConnectionFailed(self, connector, reason):
        self.link.logger.info("Failed to connect")
        reactor.callLater(1, self.reconnect, connector)

    def clientConnectionLost(self, connector, unused_reason):
        self.link.logger.info("Connection lost")
        reactor.callLater(1, self.reconnect, connector)

    def reconnect(self, connector):
        if not self.link.handshake.run_handshake():
            if self.cooldown <= 60:
                self.cooldown += 1
            reactor.callLater(self.cooldown, self.reconnect, connector)
            return
        self.reset_url()
        self.retry(connector)

    def reset_url(self):
        websocket_uri, url, port = self.link.get_url()
        self.url = websocket_uri
        (self.isSecure, self.host, self.port, self.resource, self.path, self.params) = parse_url(websocket_uri)


class DSAWebSocket(WebSocketClientProtocol):
    """
    Autobahn implementation for DSA communications.
    """
    def __init__(self):
        """
        Constructor for DSAWebSocket.
        """
        super(DSAWebSocket, self).__init__()
        self.msg = 0
        self.logger = logging.getLogger("DSLink")
        self.link = DSAWebSocket.link
        self.link.wsp = self

    def sendPingMsg(self):
        """
        Send a blank object for a ping.
        """
        if self.link.active:
            self.logger.debug("Ping")
            # noinspection PyTypeChecker
            self.sendMessage({})

    def onOpen(self):
        """
        WebSocket open event.
        """
        self.link.active = True
        self.logger.info("WebSocket Connection Established")
        task.LoopingCall(self.sendPingMsg).start(self.link.config.ping_time)

    def onClose(self, wasClean, code, reason):
        """
        WebSocket close event.
        :param wasClean: True if the close was clean.
        :param code: Close code.
        :param reason: Close reason.
        """
        self.link.active = False
        self.logger.info("WebSocket Connection Lost")

    def onMessage(self, payload, isBinary):
        """
        WebSocket message event.
        :param payload: Data to send.
        :param isBinary: True if the message is binary.
        """
        self.logger.debug("Received data: %s" % payload.decode("utf-8"))
        i = json.loads(payload.decode("utf-8"))
        m = {}
        ack = False
        if "requests" in i and len(i["requests"]) > 0:
            ack = True
            m["responses"] = self.handleRequests(i["requests"])
        if "responses" in i and len(i["responses"]) > 0:
            ack = True
            self.handleResponses(i["responses"])
        if ack:
            if "msg" in i:
                m["ack"] = i["msg"]
            # noinspection PyTypeChecker
            self.sendMessage(m)

    def handleRequests(self, requests):
        """
        Handle incoming requests.
        :param requests: List of requests.
        :return: Outgoing responses.
        """
        i = []
        for request in requests:
            i.append(Request(request, self.link).process().get_stream())
        return i

    def handleResponses(self, responses):
        """
        Handle incoming responses.
        :param responses: List of responses.
        """
        for response in responses:
            Response(response)
            if response["rid"] is 0:
                for update in response["updates"]:
                    if isinstance(update, list):
                        sid = update[0]
                        val = update[1]
                        time = update[2]
                        self.link.requester.subscription_manager.run_callback(sid, val, time)
                    elif isinstance(update, dict):
                        def get_value(vlist, key, default):
                            if key in vlist:
                                return vlist[key]
                            return default
                        sid = update["sid"]
                        val = update["value"]
                        time = update["ts"]
                        count = get_value(update, "count", 0)
                        sum = get_value(update, "sum", 0)
                        min = get_value(update, "min", 0)
                        max = get_value(update, "max", 0)
                        self.link.requester.subscription_manager.run_callback(sid, val, time, count, sum, min, max)
            elif response["rid"] in self.link.requester.request_manager.requests:
                self.link.requester.request_manager.invoke_request(response["rid"], response)

    def sendMessage(self, payload, isBinary=False, fragmentSize=None, sync=False, doNotCompress=False):
        """
        Send a message over the WebSocket.
        :param payload: Message to send.
        """
        payload["msg"] = self.msg
        self.msg += 1
        payload = json.dumps(payload, sort_keys=True, cls=JsonEncoder)
        self.logger.debug("Sent data: %s" % payload)
        payload = payload.encode("utf-8")
        super(DSAWebSocket, self).sendMessage(payload, isBinary, fragmentSize, sync, doNotCompress)



from .Response import *


class Requester:
    def __init__(self, link):
        self.link = link
        self.request_manager = RequestManager()
        self.subscription_manager = RemoteSubscriptionManager()
        self.rid = 1
        self.sid = 0

    def get_next_rid(self):
        """
        Get the next rid in the sequence. Initially starts from 1.
        :return: Next rid in sequence.
        """
        r = self.rid
        self.rid += 1
        return r

    def get_next_sid(self):
        """
        Get the next sid in the sequence. Initially starts from 0.
        :return: Next sid in the sequence.
        """
        s = self.sid
        self.sid += 1
        return s

    def list(self, path, callback):
        """
        List a remote node.
        :param path: Request path.
        :param callback: Response callback.
        """
        rid = self.get_next_rid()
        self.link.wsp.sendMessage({
            "requests": [
                {
                    "rid": rid,
                    "method": "list",
                    "path": path
                }
            ]
        }, self)
        self.request_manager.start_request(rid, "list", callback, metadata={
            "path": path
        })

    def set(self, path, value, permit=None, callback=None):
        """
        Set a remote value.
        :param path: Path of value.
        :param value: Value to set.
        :param permit: Maximum permission of set.
        :param callback: Response callback.
        """
        rid = self.get_next_rid()
        i = {
            "rid": rid,
            "method": "set",
            "path": path,
            "value": value
        }
        if permit is not None:
            i["permit"] = permit
        self.link.wsp.sendMessage({
            "requests": [
                i
            ]
        })
        if callback:
            self.request_manager.start_request(rid, "set", callback)

    def remove(self, path, callback=None):
        """
        Remove a remote value.
        :param path: Path of value.
        :param callback: Response callback.
        """
        rid = self.get_next_rid()
        self.link.wsp.sendMessage({
            "rid": rid,
            "method": "remove",
            "path": path
        })
        if callback:
            self.request_manager.start_request(rid, "remove", callback)

    def invoke(self, path, permit=None, params=None, callback=None):
        """
        Invoke a remote method.
        :param path: Path of node.
        :param permit: Maximum permission of invoke.
        :param params: Parameters of invoke.
        :param callback: Response callback.
        """
        rid = self.get_next_rid()
        i = {
            "rid": rid,
            "method": "invoke",
            "path": path
        }
        if permit is not None:
            i["permit"] = permit
        if params is not None:
            i["params"] = params
        self.link.wsp.sendMessage({
            "requests": [
                i
            ]
        })
        if callback:
            self.request_manager.start_request(rid, "invoke", callback)

    def subscribe(self, path, callback, qos=0):
        """
        Subscribe to a remote Node.
        :param path: Path of Node.
        :param callback: Response callback.
        :param qos: Quality of Service.
        """
        if type(path) is not str:
            raise ValueError("Unsupported path type.")
        rid = self.get_next_rid()
        sid = self.get_next_sid()
        self.subscription_manager.subscribe(sid, path, callback, qos)
        self.link.wsp.sendMessage({
            "requests": [
                {
                    "rid": rid,
                    "method": "subscribe",
                    "paths": [
                        {
                            "path": path,
                            "sid": sid,
                            "qos": qos
                        }
                    ]
                }
            ]
        })

    def unsubscribe(self, path):
        """
        Unsubscribe to a remote Node.
        :param path: Path of Node.
        """
        if type(path) is str:
            sid = self.subscription_manager.get_sid_by_path(path)
            self.subscription_manager.unsubscribe(sid)
            rid = self.get_next_rid()
            self.link.wsp.sendMessage({
                "requests": [
                    {
                        "rid": rid,
                        "method": "unsubscribe",
                        "sids": [
                            sid
                        ]
                    }
                ]
            })
        else:
            raise ValueError("Unsupported path type.")

    def close(self, rid):
        """
        Close a stream.
        :param rid: ID of request.
        """
        self.link.wsp.sendMessage({
            "requests": [
                {
                    "rid": rid,
                    "method": "close"
                }
            ]
        })


class RequestManager:
    """
    Manages outgoing requests and callbacks.
    """

    def __init__(self):
        """
        Constructor of RequestManager.
        """
        self.requests = {}

    def start_request(self, rid, type, callback, metadata=None):
        """
        Start a Request.
        :param rid: RID of Request.
        :param type: Type of Request.
        :param callback: Callback to invoke.
        :param metadata: Metadata for request.
        """
        if not metadata:
            metadata = {}
        self.requests[rid] = {
            "type": type,
            "callback": callback,
            "metadata": metadata
        }

    def stop_request(self, rid):
        """
        Stop a Request.
        :param rid: RID of Request.
        """
        del self.requests[rid]

    def invoke_request(self, rid, data):
        """
        Invoke a Request.
        :param rid: RID of Request.
        :param data: Data of Request.
        """
        request = self.requests[rid]
        response = data
        if request["type"] == "list":
            response = ListResponse(data, request["metadata"]["path"])
        request["callback"](response)


class RemoteSubscriptionManager:
    def __init__(self):
        self.subscriptions = {}

    def subscribe(self, sid, path, callback, qos=0):
        """
        Subscribe to a Node.
        :param sid: Subscription ID.
        :param path: Path of Node.
        :param callback: Callback to run with value.
        :param qos: Quality of Service.
        """
        if not hasattr(callback, "__call__"):
            raise ValueError("Passed callback is not callable.")
        elif sid in self.subscriptions:
            raise ValueError("Sid %s is already in use." % sid)
        self.subscriptions[sid] = {
            "path": path,
            "callback": callback,
            "qos": qos
        }

    def unsubscribe(self, sid):
        """
        Unsubscribe from a Node.
        :param sid: Subscription ID.
        """
        if sid not in self.subscriptions:
            raise ValueError("Sid %s not in use." % sid)
        del self.subscriptions[sid]

    def get_sid_by_path(self, path):
        """
        Get a Subscription ID by path.
        :param path: Path of Node.
        :return: Subscription ID.
        """
        for sid in self.subscriptions:
            meta = self.subscriptions[sid]
            if meta["path"] == path:
                return sid

    def run_callback(self, sid, value, time, count=0, sum=0, min=0, max=0):
        """
        Run a subscription's callback.
        :param sid: Subscription ID.
        :param value: Value.
        :param time: Updated at time.
        :param count: Number of values merged.
        :param sum: Sum of values merged.
        :param min: Minimum value of values merged.
        :param max: Maximum value of values merged.
        """
        self.subscriptions[sid]["callback"]((value, time, count, sum, min, max))



import base64
import hashlib
import os.path
import pickle
import pyelliptic


class Keypair:
    """
    Class to handle keypair generation, loading, and saving.
    """

    def __init__(self, location):
        """
        Keypair Constructor.
        """
        self.location = location
        self.keypair = None
        if not os.path.isfile(self.location):
            self.generate_key()
            self.save_keys()
        else:
            self.load_keys()
        sha = hashlib.sha256(self.keypair.get_pubkey())
        self.b64 = base64.urlsafe_b64encode(sha.digest()).decode("utf-8").replace("=", "")
        self.encoded_public = base64.urlsafe_b64encode(self.keypair.get_pubkey()).decode("utf-8").replace("=", "")

    def generate_key(self):
        """
        Generate a key.
        """
        self.keypair = pyelliptic.ECC(curve="prime256v1")

    def load_keys(self):
        """
        Load the keys from a file.
        """
        file = open(self.location, "rb")
        try:
            keys = pickle.load(file)
        except ValueError:
            raise ValueError("Could not load serialized keys. Possibly a Python version mismatch. "
                             "Try deleting your keys and try running again.")
        self.keypair = pyelliptic.ECC(curve="prime256v1",
                                      pubkey_x=keys["pubkey_x"],
                                      pubkey_y=keys["pubkey_y"],
                                      raw_privkey=keys["privkey"])
        file.close()

    def save_keys(self):
        """
        Save the keys to a file.
        """
        file = open(self.location, "wb")
        pickle.dump({
            "pubkey_x": self.keypair.pubkey_x,
            "pubkey_y": self.keypair.pubkey_y,
            "privkey": self.keypair.privkey
        }, file)
        file.close()
